uniform sampler2D iChannel0;
uniform sampler2D tex0;
uniform sampler2D tex1;
uniform vec2 iResolution;
uniform float iGlobalTime;
uniform float cameraRotation_x;
uniform float cameraRotation_y;
uniform float cameraRotation_z;
uniform float cameraPosition_x;
uniform float cameraPosition_y;
uniform float cameraPosition_z;
uniform float obj_rot_x;
uniform float obj_rot_y;
uniform float obj_rot_z;
uniform float logo_width;
uniform float logo_height;
uniform float fft;
uniform float fstr_val;
uniform float fft_mult;
uniform float prism_weight;
uniform float cube_reveal;
uniform float tunnel_wrip0;
uniform float tunnel_wrip1;

float time = iGlobalTime;
vec2 resolution = iResolution;

const int MAX_ITER = 64;
const float MAX_DIST = 32.0;
const float EPSILON = 0.01;
const int MAX_RECURSE = 1; // Reflections

// Rotations

mat3 rotateX(float v){return mat3(1,0,0,0,cos(v),-sin(v),0,sin(v),cos(v));}
mat3 rotateY(float v){return mat3(cos(v),0,sin(v),0,1,0,-sin(v),0,cos(v));}
mat3 rotateZ(float v){return mat3(cos(v),-sin(v),0,sin(v),cos(v),0,0,0,1);}

// Hash without Sine
// Created by David Hoskins.
#define ITERATIONS 4

// *** Use this for integer stepped ranges, ie Value-Noise/Perlin noise functions.
#define HASHSCALE1 .1031
#define HASHSCALE3 vec3(.1031, .1030, .0973)
#define HASHSCALE4 vec4(1031, .1030, .0973, .1099)

//----------------------------------------------------------------------------------------
//  1 out, 1 in...
float hash11(float p)
{
  vec3 p3  = fract(vec3(p) * HASHSCALE1);
    p3 += dot(p3, p3.yzx + 19.19);
    return fract((p3.x + p3.y) * p3.z);
}

//----------------------------------------------------------------------------------------
//  1 out, 2 in...
float hash12(vec2 p)
{
  vec3 p3  = fract(vec3(p.xyx) * HASHSCALE1);
    p3 += dot(p3, p3.yzx + 19.19);
    return fract((p3.x + p3.y) * p3.z);
}

//----------------------------------------------------------------------------------------
//  2 out, 1 in...
vec2 hash21(float p)
{
  vec3 p3 = fract(vec3(p) * HASHSCALE3);
  p3 += dot(p3, p3.yzx + 19.19);
  return fract(vec2((p3.x + p3.y)*p3.z, (p3.x+p3.z)*p3.y));
}

// END Hash without Sine

// HG_SDF, SEE README
// Maximum/minumum elements of a vector
float vmax(vec3 v) {
  return max(max(v.x, v.y), v.z);
}

// Repeat in three dimensions
vec3 pMod3(inout vec3 p, vec3 size) {
  vec3 c = floor((p + size*0.5)/size);
  p = mod(p + size*0.5, size) - size*0.5;
  return c;
}

// Repeat the domain only in positive direction. Everything in the negative half-space is unchanged.
float pModSingle1(inout float p, float size) {
  float halfsize = size*0.5;
  float c = floor((p + halfsize)/size);
  if (p >= 0.0) {
    p = mod(p + halfsize, size) - halfsize;
  }
  return c;
}

// The "Round" variant uses a quarter-circle to join the two objects smoothly:
float fOpUnionRound(float a, float b, float r) {
  vec2 u = max(vec2(r - a,r - b), vec2(0));
  return max(r, min (a, b)) - length(u);
}

// The "Chamfer" flavour makes a 45-degree chamfered edge (the diagonal of a square of size <r>):
float fOpUnionChamfer(float a, float b, float r) {
  return min(min(a, b), (a - r + b)*sqrt(0.5));
}

// Intersection has to deal with what is normally the inside of the resulting object
// when using union, which we normally don't care about too much. Thus, intersection
// implementations sometimes differ from union implementations.
float fOpIntersectionChamfer(float a, float b, float r) {
  return max(max(a, b), (a + r + b)*sqrt(0.5));
}

// Difference can be built from Intersection or Union:
float fOpDifferenceChamfer (float a, float b, float r) {
  return fOpIntersectionChamfer(a, -b, r);
}

// Primitives

float fSphere(vec3 p, float r) {
  return length(p) - r;
}

// Box: correct distance to corners
float fBox(vec3 p, vec3 b) {
  vec3 d = abs(p) - b;
  return length(max(d, vec3(0))) + vmax(min(d, vec3(0)));
}

// Cylinder standing upright on the xz plane
float fCylinder(vec3 p, float r, float height) {
  float d = length(p.xz) - r;
  d = max(d, abs(p.y) - height);
  return d;
}


// Cone with correct distances to tip and base circle. Y is up, 0 is in the middle of the base.
float fCone(vec3 p, float radius, float height) {
  vec2 q = vec2(length(p.xz), p.y);
  vec2 tip = q - vec2(0, height);
  vec2 mantleDir = normalize(vec2(height, radius));
  float mantle = dot(tip, mantleDir);
  float d = max(mantle, -q.y);
  float projected = dot(tip, vec2(mantleDir.y, -mantleDir.x));

  // distance to tip
  if ((q.y > height) && (projected < 0.0)) {
    d = max(d, length(tip));
  }

  // distance to base ring
  if ((q.x > radius) && (projected > length(vec2(height, radius)))) {
    d = max(d, length(q - vec2(radius, 0.0)));
  }
  return d;
}

// END HG_SDF, SEE README

float opDisplace( vec3 p, float obj )
{
    float d1 = obj;
    float d2 = sin(time*p.x)*sin(time*p.y)*sin(time*p.z);
    return d1+d2;
}

// Distance field: insert objects here
float distfunc(vec3 pos)
{
  float field;
  // Re-center the camera
  pos += vec3(1.0, 0.0, -1.0);


  // rotations
  pos *= rotateY(0.3*sin(cameraRotation_y));
  pos *= rotateZ(0.2*sin(0.8*cameraRotation_z) + obj_rot_z * pos.z + 0.01*cameraRotation_z);
  pos *= rotateX(0.5*sin(0.9*cameraRotation_x));

  // Save neutral space before movement
  vec3 neutral = pos;
  // Zoom!
  pos.x += 3.0*cameraPosition_z;
  // Camera movements
  pos.x -= cameraPosition_x;


  // Constant space, we'll zoom through
  vec3 obj1Space = pos;
  pMod3(obj1Space, vec3(2+prism_weight));

  // Mending
  float obj1 = fBox(obj1Space, vec3(0.1+0.012*obj_rot_x));
  float obj2 = fSphere(obj1Space, 0.2+0.012*obj_rot_x);

  obj1 = fOpDifferenceChamfer(obj1, obj2, 0.1);

  // Center obj
  vec3 shippos = neutral;
  float centerSize = 0.2+cube_reveal;
  shippos.z += 2.0;
  shippos.x -= 1.0;
  shippos *= rotateY(time);
  shippos *= rotateZ(time);
  float ship = fBox(shippos, vec3(centerSize));

  vec3 ship2pos = shippos;
  ship2pos *= rotateY(2.0*time);
  ship2pos *= rotateZ(2.0*time);
  float ship2 = fBox(ship2pos, vec3(centerSize));

  vec3 shipsphp = shippos;
  // centerSize = 0.2
  float shipsph = fSphere(shipsphp, 0.007*obj_rot_y);
  // centerSize = 0.1
  //float shipsph = fSphere(shipsphp, 0.0028*time);

  ship = fOpUnionRound(ship, ship2, 0.1);
  ship = fOpDifferenceChamfer(ship, shipsph, 0.1);
  // end center obj

  // Combine objects
  field = 1.0;
  field = min(field, obj1);
  field = min(field, ship);

  return field;
}

// Raymarch
// returns pos ended up at + normal
void rayMarch(inout vec3 pos, vec3 rayDir, out vec3 normal) {
  float totalDist = 0.0;
  float dist = EPSILON;
  for (int i = 0; i < MAX_ITER; i++)
  {
    dist = distfunc(pos);
    totalDist += dist;
    pos += dist * rayDir;

    if (dist < EPSILON || totalDist > MAX_DIST)
    break;
  }

  // Calculate normals
  if (dist < EPSILON)
  {
    vec2 eps = vec2(0.0, EPSILON);
    normal = normalize(vec3(
      distfunc(pos + eps.yxx) - distfunc(pos - eps.yxx),
      distfunc(pos + eps.xyx) - distfunc(pos - eps.xyx),
      distfunc(pos + eps.xxy) - distfunc(pos - eps.xxy)));
  }
  else {
    normal = vec3(0.0);
  }

}

// Main drawing loop
void main(){
  // Camera setup
  vec3 cameraOrigin = vec3(0.0, 0.0, 1.0);
  vec3 cameraTarget = vec3(0.0, 0.0, 0.0);
  vec3 upDirection = vec3(0.0, 1.0, 0.0);
  vec3 cameraDir = normalize(cameraTarget - cameraOrigin);
  vec3 cameraRight = normalize(cross(upDirection, cameraOrigin));
  vec3 cameraUp = cross(cameraDir, cameraRight);
  vec2 screenPos = -1.0 + 2.0 * gl_FragCoord.xy / resolution.xy; // screenPos can range from -1 to 1
  screenPos.x *= resolution.x / resolution.y;                   // Correct aspect ratio
  vec3 rayDir = normalize(cameraRight * screenPos.x + cameraUp * screenPos.y + cameraDir);

  // Call raymarch from camera origin
  vec3 pos = cameraOrigin;
  vec3 normal, color, defaultColor;
  float dist;
  defaultColor = vec3(0.153, 0.157, 0.133); // dark warm grey
  //defaultColor = vec3(0.0); // black

  // Iterate
  for (int i = 0; i < MAX_RECURSE; i++) {
    // March from camera
    rayMarch(pos, rayDir, normal);

    float diffuse, specular, distDarken;
    dist = distance(cameraOrigin, pos);

    // Lighting, from camera
    diffuse = max(0.0, dot(-rayDir, normal));
    specular = 0.1*pow(diffuse, 32.0);
    distDarken = 0.005*dist*dist;

    if (i==0) {
      // first, primary iteration
      color = defaultColor + vec3(max(0.0, diffuse + specular - distDarken));
    } else {
    // latter iterations; reflections
      color -= 0.2*vec3(max(0.0, diffuse + specular - distDarken));
    }

    // Change direction for re-march.
    pos -= rayDir*0.1;
    // Remember to step out a bit, to not hit the object we bounced off!
    rayDir = reflect(rayDir, normal);
  }

  // POST PROCESSING

  // vignette
  vec2 uv = gl_FragCoord.xy / resolution.xy-vec2(.5);
  float vignette = exp(-4.0*(uv.x*uv.x+uv.y*uv.y));
  // grain
  vec2 position = gl_FragCoord.xy;
  float a = 0.0, b = a;
    for (int t = 0; t < ITERATIONS; t++) {
        float v = float(t+1)*.152;
        vec2 pos = (position * v + time * 500.);
        a += hash12(pos);
    }
  vec3 noisecol = vec3(mix(b, a, step(uv.x, .5))) / float(ITERATIONS);
  float grain = noisecol.x + 0.5;

  vec4 finalcolor1 = vec4(color * vignette * grain, 1.0);
  finalcolor1.r = min(max(finalcolor1.r, 0.0), 1.0);
  finalcolor1 = vec4(texture2D(tex0, vec2(finalcolor1.r,0.0)).rgb,1.);
  vec4 finalcolor2 = vec4(color * vignette * grain, 1.0);
  finalcolor2.r = min(max(finalcolor2.r, 0.0), 1.0);
  finalcolor2 = vec4(texture2D(tex1, vec2(finalcolor2.r,0.0)).rgb,1.);
  gl_FragColor = logo_width * finalcolor1 + (1.0 - logo_width) * finalcolor1.rrra;

}